import datetime

from tests.utils.tests_utils import test_function

from geoformat.processing.data.union import (
    union_metadata,
    union_geolayer
)

from geoformat.conf.error_messages import (
    metadata_fields_not_same,
    metadata_geometry_crs
)
from tests.data.fields_metadata import geolayer_data_fields_metadata_complete

from tests.data.metadata import metadata_fr_dept_data_and_geometry, metadata_paris_velib

from tests.data.geolayers import (
    geolayer_btc_price_sample,
    geolayer_btc_price_sample_a,
    geolayer_btc_price_sample_b,
    geolayer_btc_price_sample_c,
)

union_metadata_parameters = {
    0: {
        "metadata_a": metadata_fr_dept_data_and_geometry,
        "metadata_b": metadata_fr_dept_data_and_geometry,
        "metadata_name": metadata_fr_dept_data_and_geometry["name"],
        "feature_serialize": None,
        "return_value": metadata_fr_dept_data_and_geometry,
    },
    1: {
        "metadata_a": metadata_fr_dept_data_and_geometry,
        "metadata_b": metadata_fr_dept_data_and_geometry,
        "metadata_name": "test_metadata",
        "feature_serialize": None,
        "return_value": {
            "name": "test_metadata",
            "fields": {
                "CODE_DEPT": {"type": "String", "width": 2, "index": 0},
                "NOM_DEPT": {"type": "String", "width": 23, "index": 1},
            },
            "geometry_ref": {"type": {"Polygon", "MultiPolygon"}, "crs": 2154},
        },
    },
    2: {
        "metadata_a": metadata_fr_dept_data_and_geometry,
        "metadata_b": metadata_paris_velib,
        "metadata_name": "test_metadata",
        "feature_serialize": None,
        "return_value": metadata_fields_not_same,
    },
    3: {
        "metadata_a": {
            "name": "FRANCE_DPT_GENERALIZE_LAMB93_ROUND_DATA_AND_GEOMETRY",
            "fields": geolayer_data_fields_metadata_complete,
            "geometry_ref": {"type": {"MultiPolygon"}, "crs": 2154},
        },
        "metadata_b": {
            "name": "FRANCE_DPT_GENERALIZE_LAMB93_ROUND_DATA_AND_GEOMETRY",
            "fields": geolayer_data_fields_metadata_complete,
            "geometry_ref": {"type": {"Polygon"}, "crs": 2154},
        },
        "metadata_name": "test_metadata",
        "feature_serialize": None,
        "return_value": {
            "name": "test_metadata",
            "fields": geolayer_data_fields_metadata_complete,
            "geometry_ref": {"type": {"MultiPolygon", "Polygon"}, "crs": 2154},
        }
    },
    4: {
        "metadata_a": {
            "name": "FRANCE_DPT_GENERALIZE_LAMB93_ROUND_DATA_AND_GEOMETRY",
            "fields": geolayer_data_fields_metadata_complete,
            "geometry_ref": {"type": {"MultiPolygon"}, "crs": 2154},
        },
        "metadata_b": {
            "name": "FRANCE_DPT_GENERALIZE_LAMB93_ROUND_DATA_AND_GEOMETRY",
            "fields": geolayer_data_fields_metadata_complete,
            "geometry_ref": {"type": {"Polygon"}, "crs": 4326},
        },
        "metadata_name": "test_metadata",
        "feature_serialize": None,
        "return_value": metadata_geometry_crs
    },
    5: {
        "metadata_a": {'name': 'BTC_DAILY_PRICE', 'fields': {'DATE': {'type': 'Date', 'index': 0}, 'DAYS': {'type': 'Integer', 'index': 1}, 'TIMESTAMP': {'type': 'Integer', 'index': 2}, 'USD_PRICE_CLOSE': {'type': 'Real', 'width': 7, 'precision': 2, 'index': 3}, 'PRICE_ESTIMATE_MINUS_1STD': {'type': 'Real', 'width': 22, 'precision': 20, 'index': 4}, 'PRICE_ESTIMATE_MINUS_2STD': {'type': 'Real', 'width': 21, 'precision': 20, 'index': 5}, 'PRICE_ESTIMATE_2019': {'type': 'Real', 'width': 22, 'precision': 20, 'index': 6}, 'PRICE_ESTIMATE': {'type': 'Real', 'width': 22, 'precision': 20, 'index': 7}, 'PRICE_ESTIMATE_PLUS_1STD': {'type': 'Real', 'width': 22, 'precision': 20, 'index': 8}, 'PRICE_ESTIMATE_PLUS_2STD': {'type': 'Real', 'width': 22, 'precision': 20, 'index': 9}}},
        "metadata_b": {'name': 'BTC_DAILY_PRICE', 'fields': {'DATE': {'type': 'Date', 'index': 0}, 'DAYS': {'type': 'Integer', 'index': 1}, 'TIMESTAMP': {'type': 'Integer', 'index': 2}, 'USD_PRICE_CLOSE': {'type': 'Real', 'width': 7, 'precision': 2, 'index': 3}, 'PRICE_ESTIMATE_MINUS_1STD': {'type': 'Real', 'width': 18, 'precision': 15, 'index': 4}, 'PRICE_ESTIMATE_MINUS_2STD': {'type': 'Real', 'width': 18, 'precision': 15, 'index': 5}, 'PRICE_ESTIMATE_2019': {'type': 'Real', 'width': 19, 'precision': 15, 'index': 6}, 'PRICE_ESTIMATE': {'type': 'Real', 'width': 17, 'precision': 14, 'index': 7}, 'PRICE_ESTIMATE_PLUS_1STD': {'type': 'Real', 'width': 17, 'precision': 14, 'index': 8}, 'PRICE_ESTIMATE_PLUS_2STD': {'type': 'Real', 'width': 18, 'precision': 14, 'index': 9}}},
        "metadata_name": "BTC_DAILY_PRICE",
        "feature_serialize": None,
        "return_value": {'name': 'BTC_DAILY_PRICE', 'fields': {'DATE': {'type': 'Date', 'index': 0}, 'DAYS': {'type': 'Integer', 'index': 1}, 'TIMESTAMP': {'type': 'Integer', 'index': 2}, 'USD_PRICE_CLOSE': {'type': 'Real', 'width': 7, 'precision': 2, 'index': 3}, 'PRICE_ESTIMATE_MINUS_1STD': {'type': 'Real', 'width': 22, 'precision': 20, 'index': 4}, 'PRICE_ESTIMATE_MINUS_2STD': {'type': 'Real', 'width': 21, 'precision': 20, 'index': 5}, 'PRICE_ESTIMATE_2019': {'type': 'Real', 'width': 22, 'precision': 20, 'index': 6}, 'PRICE_ESTIMATE': {'type': 'Real', 'width': 22, 'precision': 20, 'index': 7}, 'PRICE_ESTIMATE_PLUS_1STD': {'type': 'Real', 'width': 22, 'precision': 20, 'index': 8}, 'PRICE_ESTIMATE_PLUS_2STD': {'type': 'Real', 'width': 22, 'precision': 20, 'index': 9}}}
    },
}

union_geolayer_parameters = {
    0: {
        "geolayer_list": [geolayer_btc_price_sample_a, geolayer_btc_price_sample_b, geolayer_btc_price_sample_c],
        "geolayer_name": 'BTC_DAILY_PRICE',
        "serialize": False,
        "return_value": {'metadata': {'name': 'BTC_DAILY_PRICE', 'fields': {'DATE': {'type': 'Date', 'index': 0}, 'DAYS': {'type': 'Integer', 'index': 1}, 'TIMESTAMP': {'type': 'Integer', 'index': 2}, 'USD_PRICE_CLOSE': {'type': 'Real', 'width': 7, 'precision': 2, 'index': 3}}}, 'features': {0: {'attributes': {'DATE': datetime.date(2022, 4, 27), 'DAYS': 4865, 'TIMESTAMP': 1651010400, 'USD_PRICE_CLOSE': 39252.04}}, 1: {'attributes': {'DATE': datetime.date(2022, 4, 28), 'DAYS': 4866, 'TIMESTAMP': 1651096800, 'USD_PRICE_CLOSE': 39749.75}}, 2: {'attributes': {'DATE': datetime.date(2022, 4, 29), 'DAYS': 4867, 'TIMESTAMP': 1651183200, 'USD_PRICE_CLOSE': 38594.22}}, 3: {'attributes': {'DATE': datetime.date(2022, 4, 30), 'DAYS': 4868, 'TIMESTAMP': 1651269600, 'USD_PRICE_CLOSE': 37650.13}}, 4: {'attributes': {'DATE': datetime.date(2022, 5, 1), 'DAYS': 4869, 'TIMESTAMP': 1651356000, 'USD_PRICE_CLOSE': 38480.53}}, 5: {'attributes': {'DATE': datetime.date(2022, 5, 2), 'DAYS': 4870, 'TIMESTAMP': 1651442400, 'USD_PRICE_CLOSE': 38513.01}}, 6: {'attributes': {'DATE': datetime.date(2022, 5, 3), 'DAYS': 4871, 'TIMESTAMP': 1651528800, 'USD_PRICE_CLOSE': 37725.38}}, 7: {'attributes': {'DATE': datetime.date(2022, 5, 4), 'DAYS': 4872, 'TIMESTAMP': 1651615200, 'USD_PRICE_CLOSE': 39680.21}}, 8: {'attributes': {'DATE': datetime.date(2022, 5, 5), 'DAYS': 4873, 'TIMESTAMP': 1651701600, 'USD_PRICE_CLOSE': 36546.86}}, 9: {'attributes': {'DATE': datetime.date(2022, 5, 6), 'DAYS': 4874, 'TIMESTAMP': 1651788000, 'USD_PRICE_CLOSE': 36009.89}}, 10: {'attributes': {'DATE': datetime.date(2022, 5, 7), 'DAYS': 4875, 'TIMESTAMP': 1651874400, 'USD_PRICE_CLOSE': 35469.21}}, 11: {'attributes': {'DATE': datetime.date(2022, 5, 8), 'DAYS': 4876, 'TIMESTAMP': 1651960800, 'USD_PRICE_CLOSE': 34033.77}}, 12: {'attributes': {'DATE': datetime.date(2022, 5, 9), 'DAYS': 4877, 'TIMESTAMP': 1652047200, 'USD_PRICE_CLOSE': 30076.9}}, 13: {'attributes': {'DATE': datetime.date(2022, 5, 10), 'DAYS': 4878, 'TIMESTAMP': 1652133600, 'USD_PRICE_CLOSE': 31013.12}}, 14: {'attributes': {'DATE': datetime.date(2022, 5, 11), 'DAYS': 4879, 'TIMESTAMP': 1652220000, 'USD_PRICE_CLOSE': 28590.66}}, 15: {'attributes': {'DATE': datetime.date(2022, 5, 12), 'DAYS': 4880, 'TIMESTAMP': 1652306400, 'USD_PRICE_CLOSE': 28915.72}}, 16: {'attributes': {'DATE': datetime.date(2022, 5, 13), 'DAYS': 4881, 'TIMESTAMP': 1652392800, 'USD_PRICE_CLOSE': 29244.83}}, 17: {'attributes': {'DATE': datetime.date(2022, 5, 14), 'DAYS': 4882, 'TIMESTAMP': 1652479200, 'USD_PRICE_CLOSE': 30050.31}}, 18: {'attributes': {'DATE': datetime.date(2022, 5, 15), 'DAYS': 4883, 'TIMESTAMP': 1652565600, 'USD_PRICE_CLOSE': 31296.11}}, 19: {'attributes': {'DATE': datetime.date(2022, 5, 16), 'DAYS': 4884, 'TIMESTAMP': 1652652000, 'USD_PRICE_CLOSE': 29838.5}}, 20: {'attributes': {'DATE': datetime.date(2022, 5, 17), 'DAYS': 4885, 'TIMESTAMP': 1652738400, 'USD_PRICE_CLOSE': 30415.91}}, 21: {'attributes': {'DATE': datetime.date(2022, 5, 18), 'DAYS': 4886, 'TIMESTAMP': 1652824800, 'USD_PRICE_CLOSE': None}}, 22: {'attributes': {'DATE': datetime.date(2022, 5, 19), 'DAYS': 4887, 'TIMESTAMP': 1652911200}}}}
    }
}


def test_all():
    # union_metadata
    print(test_function(union_metadata, union_metadata_parameters))

    # union geolayer
    print(test_function(union_geolayer, union_geolayer_parameters))


if __name__ == "__main__":
    test_all()



from geoformat.geoprocessing.connectors.predicates import (
    point_intersects_point,
    point_intersects_segment,
    point_intersects_bbox,
    segment_intersects_segment,
    segment_intersects_bbox,
    bbox_intersects_bbox,
    ccw_or_cw_segments,
    point_position_segment
)

from tests.utils.tests_utils import test_function

point_intersects_point_parameters = {
    0: {
        'point_a': (10, 10),
        'point_b': (10, 10),
        'tolerance': None,
        'return_value': True
    },
    1: {
        'point_a': (-5, 5),
        'point_b': (5, -5),
        'tolerance': None,
        'return_value': False
    },
    3: {
        'point_a': (10, 10),
        'point_b': (10, 10),
        'tolerance': 0.0,
        'return_value': True
    },
    4: {
        'point_a': (-5, 5),
        'point_b': (5, -5),
        'tolerance': 0.,
        'return_value': False
    },
    5: {
        'point_a': (10.1, 10.),
        'point_b': (10, 10),
        'tolerance': 0.1,
        'return_value': True
    },
    6: {
        'point_a': (-5, 5),
        'point_b': (5, -5),
        'tolerance': 0.1,
        'return_value': False
    },
    7: {
        'point_a': (10.1, 10.),
        'point_b': (10, 10),
        'tolerance': 0.01,
        'return_value': False
    },
    8: {
        'point_a': (-5, 5),
        'point_b': (5, -5),
        'tolerance': 15,
        'return_value': True
    },
}

point_intersects_segment_parameters = {
    0: {
        'point': (0, 0),
        'segment': ((-10, -10), (10, 10)),
        'tolerance': None,
        'return_value': True
    },
    1: {
        'point': (-11, -11),
        'segment': ((-10, -10), (10, 10)),
        'tolerance': None,
        'return_value': False
    },
    2: {
        'point': (2, 2),
        'segment': ((1, 2), (3, 2)),
        'tolerance': None,
        'return_value': True
    },
    3: {
        'point': (1.9, 1.9),
        'segment': ((1, 2), (3, 2)),
        'tolerance': None,
        'return_value': False
    },
    4: {
        'point': (1.9, 1.9),  # point near the middle of segment
        'segment': ((1, 2), (3, 2)),  # horizontal segment
        'tolerance': 0.1,
        'return_value': True
    },
    5: {
        'point': (1.9, 1.9),  # point near the middle of segment
        'segment': ((1, 2), (3, 2)),  # horizontal segment
        'tolerance': 0.01,
        'return_value': False
    },
    6: {
        'point': (1.99, 1.99),  # point near the middle of segment
        'segment': ((1, 2), (3, 2)),  # horizontal segment
        'tolerance': 0.01,
        'return_value': True
    },
    7: {
        'point': (0.9, 1.9),  # point near the beginning of segment
        'segment': ((1, 2), (3, 2)),  # horizontal segment
        'tolerance': 0.1,
        'return_value': True
    },
    8: {
        'point': (0.9, 1.9),  # point near the beginning of segment
        'segment': ((1, 2), (3, 2)),  # horizontal segment
        'tolerance': 0.01,
        'return_value': False
    },
    9: {
        'point': (0.99, 1.99),  # point near the beginning of segment
        'segment': ((1, 2), (3, 2)),  # horizontal segment
        'tolerance': 0.01,
        'return_value': True
    },
    10: {
        'point': (3.1, 2.1),  # point near the ending of segment
        'segment': ((1, 2), (3, 2)),  # horizontal segment
        'tolerance': 0.1,
        'return_value': True
    },
    11: {
        'point': (3.1, 2.1),  # point near the ending of segment
        'segment': ((1, 2), (3, 2)),  # horizontal segment
        'tolerance': 0.01,
        'return_value': False
    },
    12: {
        'point': (3.01, 2.01),  # point near the ending of segment
        'segment': ((1, 2), (3, 2)),  # horizontal segment
        'tolerance': 0.01,
        'return_value': True
    },
    13: {
        'point': (-2.1, -1.9),  # point near the middle of segment
        'segment': ((-2, -1), (-2, -3)),  # vertical segment
        'tolerance': 0.1,
        'return_value': True
    },
    14: {
        'point': (-2.1, -1.9),  # point near the middle of segment
        'segment': ((-2, -1), (-2, -3)),  # vertical segment
        'tolerance': 0.01,
        'return_value': False
    },
    15: {
        'point': (-2.01, -1.09),  # point near the middle of segment
        'segment': ((-2, -1), (-2, -3)),  # vertical segment
        'tolerance': 0.01,
        'return_value': True
    },
    16: {
        'point': (-1.9, -0.9),  # point near at the segment beginning
        'segment': ((-2, -1), (-2, -3)),  # vertical segment
        'tolerance': 0.1,
        'return_value': True
    },
    17: {
        'point': (-1.9, -0.9),  # point near at the segment beginning
        'segment': ((-2, -1), (-2, -3)),  # vertical segment
        'tolerance': 0.01,
        'return_value': False
    },
    18: {
        'point': (-1.99, -0.99),  # point near at the segment beginning
        'segment': ((-2, -1), (-2, -3)),  # vertical segment
        'tolerance': 0.01,
        'return_value': True
    },
    19: {
        'point': (-2.1, -2.9),  # point near at the segment ending
        'segment': ((-2, -1), (-2, -3)),  # vertical segment
        'tolerance': 0.1,
        'return_value': True
    },
    20: {
        'point': (-2.1, -2.9),  # point near at the segment ending
        'segment': ((-2, -1), (-2, -3)),  # vertical segment
        'tolerance': 0.01,
        'return_value': False
    },
    21: {
        'point': (-2.01, -2.99),  # point near at the segment ending
        'segment': ((-2, -1), (-2, -3)),  # vertical segment
        'tolerance': 0.01,
        'return_value': True
    },
    22: {
        'point': (-0.1, 4),  # point near the middle of segment
        'segment': ((-1, 3), (1, 5)),
        'tolerance': 0.1,
        'return_value': True
    },
    23: {
        'point': (-0.1, 4),  # point near the middle of segment
        'segment': ((-1, 3), (1, 5)),
        'tolerance': 0.01,
        'return_value': False
    },
    24: {
        'point': (-0.01, 4),  # point near the middle of segment
        'segment': ((-1, 3), (1, 5)),
        'tolerance': 0.01,
        'return_value': True
    },
    25: {
        'point': (1.99999, 2.00001),
        'segment': ((1, 1), (3, 3)),
        'tolerance': 0.0001,
        'return_value': True
    },
    26: {
        'point': (1.99999, 2.00001),
        'segment': ((1, 1), (3, 3)),
        'tolerance': 0.00001,
        'return_value': True
    },
    27: {
        'point': (2, 2.00001),
        'segment': ((1, 1), (3, 3)),
        'tolerance': 0.00001,
        'return_value': True
    },
    28: {
        'point': [645301, 6779557.0331],
        'segment': [[645285, 6779558], [647006, 6779454]],
        'tolerance': 0.00001,
        'return_value': False
    },
    29: {
        'point': [645301, 6779557.03312],
        'segment': [[645285, 6779558], [647006, 6779454]],
        'tolerance': 0.0000001,
        'return_value': False
    },
    30: {
        'point': [645301, 6779557.03312],
        'segment': [[645285, 6779558], [647006, 6779454]],
        'tolerance': 0.000001,
        'return_value': True
    },
    31: {
        'point': [645301, 6779557.03312],
        'segment': [[645285, 6779558], [647006, 6779454]],
        'tolerance': None,
        'return_value': False
    },
    32: {
        'point': [645301, 6779557.03312],
        'segment': [[645285, 6779558], [647006, 6779454]],
        'tolerance': 0.0000001,
        'return_value': False
    },
    33: {
        'point': [645301, 6779557.03312],
        'segment': [[645285, 6779558], [647006, 6779454]],
        'tolerance': 0.000001,
        'return_value': True
    },
    34: {
        "point": (2, 2.05),
        "segment": ((1.95, 2), (2.05, 2)),  # segment length < tolerance
        "tolerance": 0.05,
        "return_value": True
    },
    35: {
        "point": (2, 2.05),
        "segment": ((1.99, 2), (2.01, 2)),
        "tolerance": 0.05,
        "return_value": True
    },
    36: {
        "point": (-10, -5),
        "segment": ([-10, -10], [-10, 10]),
        "tolerance": 1,
        "return_value": True
    },
    37: {
        'point': (0, 0),
        'segment': ((-10, -10), (10, 10)),
        "tolerance": None,
        'return_value': True
    },
    38: {
        'point': (10, 10),
        'segment': ((-10, -10), (10, 10)),
        "tolerance": None,
        'return_value': True
    },
    39: {
        'point': (-10, -10),
        'segment': ((-10, -10), (10, 10)),
        "tolerance": None,
        'return_value': True
    },
    40: {
        'point': (-11, -11),
        'segment': ((-10, -10), (10, 10)),
        "tolerance": None,
        'return_value': False
    },
}


point_intersects_bbox_parameters = {
    0: {
        'point': (0, 0),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    1: {
        'point': (-10, -10),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    2: {
        'point': (-10, 10),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    3: {
        'point': (10, 10),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    4: {
        'point': (10, -10),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    5: {
        'point': (-10, 0),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    6: {
        'point': (0, 10),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    7: {
        'point': (10, 0),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    8: {
        'point': (0, -10),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    9: {
        'point': (-10, -20),
        'bbox': (-10, -10, 10, 10),
        'return_value': False
    }
}

segment_intersects_segment_parameters = {
    0: {
        'segment_a': ((10, 10), (-10, -10)),
        'segment_b': ((-10, 10), (10, -10)),
        'return_value': True
    },
    1: {
        'segment_a': ((10, 10), (-10, -10)),
        'segment_b': ((0, 10), (20, 10)),
        'return_value': True
    },
    2: {
        'segment_a': ((10, 10), (-10, -10)),
        'segment_b': ((-10, -10), (0, -10)),
        'return_value': True
    },
    3: {
        'segment_a': ((10, 10), (-10, -10)),
        'segment_b': ((10, 10), (20, 20)),
        'return_value': True
    },
    4: {
        'segment_a': ((10, 10), (-10, -10)),
        'segment_b': ((-10, -10), (-20, -20)),
        'return_value': True
    },
    5: {
        'segment_a': ((10, 10), (-10, -10)),
        'segment_b': ((3, 0), (5, -2)),
        'return_value': False
    },
}

segment_intersects_bbox_parameters = {
    0: {
        'segment': ((-2, 0), (2, 0)),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    1: {
        'segment': ((-5, -10), (5, -10)),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    2: {
        'segment': ((-10, -20), (-10, 20)),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    3: {
        'segment': ((5, 10), (-20, 10)),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    4: {
        'segment': ((10, 5), (10, 30)),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    5: {
        'segment': ((-9, 11), (-11, 9)),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    6: {
        'segment': ((-10, -10), (-20, -10)),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    7: {
        'segment': ((11, -9), (9, 11)),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    8: {
        'segment': ((10, 30), (10, 10)),
        'bbox': (-10, -10, 10, 10),
        'return_value': True
    },
    9: {
        'segment': ((20, 5), (30, 5)),
        'bbox': (-10, -10, 10, 10),
        'return_value': False
    },
    10: {
        'segment': ([-62136.394556539126, 5596571.563702659], [-60626.17017134098, 5597872.545653571]),
        'bbox': (-61500.0, 5597250.0, -61450.0, 5597300.0),
        'return_value': False
    }
}

bbox_intersects_bbox_parameters = {
    0: {
        'bbox_a': (-10, -10, 10, 10),
        'bbox_b': (-10, -10, 10, 10),
        'return_value': True
    },
    1: {
        'bbox_a': (-10, -10, 10, 10),
        'bbox_b': (-20, -20, 0, 0),
        'return_value': True
    },
    2: {
        'bbox_a': (-10, -10, 10, 10),
        'bbox_b': (-20, -20, 0, 0),
        'return_value': True
    },
    3: {
        'bbox_a': (-10, -10, 10, 10),
        'bbox_b': (-30, -30, -10, 0),
        'return_value': True
    },
    4: {
        'bbox_a': (-10, -10, 10, 10),
        'bbox_b': (11, 11, 40, 40),
        'return_value': False
    }
}

ccw_or_cw_segments_parameters = {
    0: {
        'segment_a': ((-5, -5), (5, 5)),
        'segment_b':  ((5, 5), (-1, 2)),
        'return_value': 'CCW'
    },
    1: {
        'segment_a': ((-5, -5), (5, 5)),
        'segment_b': ((5, 5), (5, -5)),
        'return_value': 'CW'
    },
    2: {
        'segment_a': ((-5, -5), (5, 5)),
        'segment_b': ((5, 5), (10, 10)),
        'return_value': 'NEITHER'
    },
    3: {
        'segment_a': ((-5, -5), (5, 5)),
        'segment_b': ((4, 4), (10, 10)),
        'return_value': None
    },
    4: {
        'segment_a': ((5, 5), (-5, -5)),
        'segment_b': ((-5, -5), (-1, 2)),
        'return_value': 'CW'
    },
    5: {
        'segment_a': ((5, 5), (-5, -5)),
        'segment_b': ((-5, -5), (5, -5)),
        'return_value': 'CCW'
    },
    6: {
        'segment_a': ((5, 5), (-5, -5)),
        'segment_b': ((-5, -5), (10, 10)),
        'return_value': 'NEITHER'
    },
    7: {
        'segment_a':  ((5, 5), (-5, -5)),
        'segment_b':  ((-4, -4), (10, 10)),
        'return_value': None
    },
}

point_position_segment_parameters = {
    0: {
        'segment': ((-5, -5), (5, 5)),
        'point':  (-1, 2),
        'return_value': 'LEFT'
    },
    1: {
        'segment': ((-5, -5), (5, 5)),
        'point':  (5, -5),
        'return_value': 'RIGHT'
    },
    2: {
        'segment': ((-5, -5), (5, 5)),
        'point': (10, 10),
        'return_value': 'NEITHER'
    },
    3: {
        'segment': ((-5, -5), (5, 5)),
        'point': (0, 0),
        'return_value': 'ON'
    },
    4: {
        'segment': ((5, 5), (-5, -5)),
        'point': (-1, 2),
        'return_value': 'RIGHT'
    },
    5: {
        'segment': ((5, 5), (-5, -5)),
        'point': (5, -5),
        'return_value': 'LEFT'
    },
    6: {
        'segment': ((5, 5), (-5, -5)),
        'point': (10, 10),
        'return_value': 'NEITHER'
    },
    7: {
        'segment': ((5, 5), (-5, -5)),
        'point': (0, 0),
        'return_value': 'ON'
    }
}


def test_all():
    # point_intersects_point
    print(test_function(point_intersects_point, point_intersects_point_parameters))

    # point_intersects_segment
    print(test_function(point_intersects_segment, point_intersects_segment_parameters))

    # point_intersects_bbox
    print(test_function(point_intersects_bbox, point_intersects_bbox_parameters))

    # segment_intersects_segment
    print(test_function(segment_intersects_segment, segment_intersects_segment_parameters))

    # segment_intersects_bbox
    print(test_function(segment_intersects_bbox, segment_intersects_bbox_parameters))

    # bbox_intersects_bbox
    print(test_function(bbox_intersects_bbox, bbox_intersects_bbox_parameters))

    # ccw_or_cw_segments
    print(test_function(ccw_or_cw_segments, ccw_or_cw_segments_parameters))

    # point_position_segment
    print(test_function(point_position_segment, point_position_segment_parameters))


if __name__ == '__main__':
    test_all()



import datetime


def format_datetime_object_to_str_value(
    datetime_value,
    format=["year", "month", "day"]
):
    """
    Convert a datetime object to a string representation based on specified components.

    This function allows selective formatting of a datetime object into a string by choosing specific components
    such as year, month, day, hour, second, and microsecond. The components are concatenated in the order
    provided in the format list.

    Parameters:
    datetime_value (datetime): The datetime object to be formatted.
    format (list of str, optional): A list of string components to include in the output.
        Default is ["year", "month", "day"]. Valid components are "year", "month", "day", "hour",
        "second", and "microsecond".

    Returns:
    str: A string representation of the datetime object based on the specified format.

    Raises:
    Exception: If an invalid component is specified in the format list.
    """

    if isinstance(format, str):
        format = [format]

    return_value = ""
    for v in format:
        if v == "year":
            return_value += str(datetime_value.year).zfill(4)
        elif v == "month":
            return_value += str(datetime_value.month).zfill(2)
        elif v == "day":
            return_value += str(datetime_value.day).zfill(2)
        elif v == "hour":
            return_value += str(datetime_value.hour).zfill(2)
        elif v == "second":
            return_value += str(datetime_value.second).zfill(2)
        elif v == "microsecond":
            return_value += str(datetime_value.microsecond).zfill(6)
        else:
            raise Exception()

    return return_value


def date_to_int(date_value, epoch=datetime.date(1970, 1, 1)):
    """
    Convert a date object to an integer representing the number of days since a specified epoch.

    This function calculates the number of days from a given epoch (default is 1970-01-01) to the specified date.
    It's useful for converting date values into a consistent numeric format for calculations or storage.

    Parameters:
    date_value (datetime.date): The date value to be converted.
    epoch (datetime.date, optional): The epoch (start date) from which the days are counted.
        Default is 1970-01-01.

    Returns:
    int: The number of days from the epoch to the given date.
    """
    return (date_value - epoch).days


def time_to_int(time_value):
    """
    Convert a time object to an integer representing the time in microseconds.

    This function transforms a time object into an integer representing the total number of microseconds
    since midnight. It's useful for time calculations or storing time values in a compact numeric format.

    Parameters:
    time_value (datetime.time): The time value to be converted.

    Returns:
    int: The number of microseconds since midnight represented by the time object.
    """
    return int(
        time_value.hour * 3600 * 1e6
        + time_value.minute * 60 * 1e6
        + time_value.second * 1e6
        + time_value.microsecond
    )

def int_to_date(int_value, epoch=datetime.date(1970, 1, 1)):
    """
    Convert an integer representing the number of days since a specified epoch to a date object.

    Parameters:
    int_value (int): The number of days since the epoch.
    epoch (datetime.date, optional): The epoch (start date) from which the days are counted.
        Default is 1970-01-01.

    Returns:
    datetime.date: The date corresponding to the given number of days since the epoch.
    """
    return epoch + datetime.timedelta(days=int_value)

def int_to_time(int_value):
    """
    Convert an integer representing time in microseconds since midnight to a time object.

    Parameters:
    int_value (int): The number of microseconds since midnight.

    Returns:
    datetime.time: The time object represented by the given number of microseconds since midnight.
    """
    seconds, int_value = divmod(int_value, 1e6)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    return datetime.time(int(hours), int(minutes), int(seconds), int(int_value))



from tests.utils.tests_utils import test_function
from geoformat.geoprocessing.geoparameters.lines import (
    get_slope_between_two_points,
    get_intercept_from_point_and_slope,
    get_slope_from_point_and_intercept,
    line_parameters,
    perpendicular_line_parameters_at_point,
    point_at_distance_with_line_parameters,
    crossing_point_from_lines_parameters
)

get_slope_between_two_points_parameters = {
    0: {
        "point_a": (0, 0),
        "point_b": (2, 2),
        "return_value": 1
    },
    1: {
        "point_a": (0, 0),
        "point_b": (2, -2),
        "return_value": -1
    },
    2: {
        "point_a": (0, 0),
        "point_b": (2, 0),
        "return_value": 0
    },
    3: {
        "point_a": (-3, -3),
        "point_b": (-3, -5),
        "return_value": 'VERTICAL'
    }
}

get_intercept_from_point_and_slope_parameters = {
    0: {
        "point": (2, 2),
        "slope": 1,
        "return_value": 0
    },
    1: {
        "point": (2, 2),
        "slope": -1,
        "return_value": 4
    },
    2: {
        "point": (-2, -2),
        "slope": 1,
        "return_value": 0
    },
    3: {
        "point": (-2, -2),
        "slope": -1,
        "return_value": -4
    },
    4: {
        "point": (3, 2),
        "slope": 3,
        "return_value": -7
    },
    5: {
        "point": (2, 2),
        "slope": 0,
        "return_value": 2
    },
    6: {
        "point": (-2, -2),
        "slope": 'VERTICAL',
        "return_value": -2
    },
}

get_slope_from_point_and_intercept_parameters = {
    0: {
        "point": (2, 2),
        "return_value": 1,
        "intercept": 0
    },
    1: {
        "point": (2, 2),
        "return_value": -1,
        "intercept": 4
    },
    2: {
        "point": (-2, -2),
        "return_value": 1,
        "intercept": 0
    },
    3: {
        "point": (-2, -2),
        "return_value": -1,
        "intercept": -4
    },
    4: {
        "point": (3, 2),
        "return_value": 3,
        "intercept": -7
    },
    5: {
        "point": (2, 2),
        "return_value": 0,
        "intercept": 2
    },
    6: {
        "point": (-2, -2),
        "return_value": 'VERTICAL',
        "intercept": False
    },
}

line_parameters_parameters = {
    0: {
        'segment': ([645285, 6779558], [647006, 6779454]),
        'return_value': {'slope': -0.06042998256827426, 'intercept': 6818552.5613015685}
    },
    1: {
        'segment': ((10, 10), (100, 20)),
        'return_value': {'slope': 0.1111111111111111, 'intercept': 8.88888888888889}
    },
    2: {
        'segment': ((0, 1), (9, 4)),
        'return_value': {'slope': 1/3, 'intercept': 1.}
    },
    3: {
        'segment': ((-5, 0), (5, 0)),
        'return_value': {'slope': 0.0, 'intercept': 0.0}
    },
    4: {
        'segment': ((0, -5), (0, 5)),
        'return_value': {'slope': 'VERTICAL', 'intercept': 0.0}
    }
}


perpendicular_line_parameters_at_point_parameters = {
    0: {
        'line_parameters': {'slope': 1/3, 'intercept': 1.},
        'point': (3, 2),
        'return_value': {'slope': -3.0, 'intercept': 11.0}
    },
    1: {
        'line_parameters': {'slope': 0.0, 'intercept': 0.0},
        'point': (0, 0),
        'return_value': {'slope': 'VERTICAL', 'intercept': 0.}
    },
    2:  {
        'line_parameters':  {'slope': 'VERTICAL', 'intercept': 5},
        'point': (0, 0),
        'return_value': {'slope': 0., 'intercept': 0.}
    }

}


point_at_distance_with_line_parameters_parameters = {
    0: {
        'start_point': (0, 0),
        'line_parameters':  {'slope': 1, 'intercept': 0},
        'distance': 5,
        'offset_distance': None,
        'return_value': (3.5355339059327373, 3.5355339059327373),
    },
    1: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 'VERTICAL', 'intercept': 0.},
        'distance': 5,
        'offset_distance': None,
        'return_value': (0, 5),
    },
    2: {
        'start_point': (0, 0),
        'line_parameters':  {'slope': 0.0, 'intercept': 0.0},
        'distance': 5,
        'offset_distance': None,
        'return_value': (5, 0),
    },
    3: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 1, 'intercept': 0},
        'distance': 5,
        'offset_distance': 2,
        'return_value': (4.949747468305832, 2.121320343559643),
    },
    4: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 'VERTICAL', 'intercept': 0.},
        'distance': 5,
        'offset_distance': 2,
        'return_value': (2, 5),
    },
    5: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 0.0, 'intercept': 0.0},
        'distance': 5,
        'offset_distance': 2,
        'return_value': (5, 2),
    },
    6: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 1, 'intercept': 0},
        'distance': -5,
        'offset_distance': 2,
        'return_value': (-2.1213203435596424, -4.949747468305832),
    },
    7: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 'VERTICAL', 'intercept': 0.},
        'distance': -5,
        'offset_distance': 2,
        'return_value': (2, -5),
    },
    8: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 0.0, 'intercept': 0.0},
        'distance': -5,
        'offset_distance': 2,
        'return_value': (-5, 2),
    },
    9: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 1, 'intercept': 0},
        'distance': -5,
        'offset_distance': -2,
        'return_value': (-4.949747468305832, -2.121320343559643),
    },
    10: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 'VERTICAL', 'intercept': 0.},
        'distance': -5,
        'offset_distance': -2,
        'return_value': (-2, -5),
    },
    11: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 0.0, 'intercept': 0.0},
        'distance': -5,
        'offset_distance': -2,
        'return_value': (-5, -2),
    },
    12: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 0.0, 'intercept': 0.0},
        'distance': 0,
        'offset_distance': 1,
        'return_value': (0, 1),
    },
    13: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 0.0, 'intercept': 0.0},
        'distance': 0,
        'offset_distance': -1,
        'return_value': (0, -1),
    },
    14: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 'VERTICAL', 'intercept': 0.0},
        'distance': 0,
        'offset_distance': 1,
        'return_value': (1, 0),
    },
    15: {
        'start_point': (0, 0),
        'line_parameters': {'slope': 'VERTICAL', 'intercept': 0.0},
        'distance': 0,
        'offset_distance': -1,
        'return_value': (-1, 0),
    }
}

crossing_point_from_lines_parameters_parameters = {
    0: {
        'line_parameter_a': {'slope': 0.0, 'intercept': 0.0},
        'line_parameter_b': {'slope': 'VERTICAL', 'intercept': 0.},
        'return_value': (0, 0),
    },
    1: {
        'line_parameter_a': {'slope': 0.0, 'intercept': -2},
        'line_parameter_b': {'slope': 'VERTICAL', 'intercept': -3.},
        'return_value': (-3, -2),
    },
    2: {
        'line_parameter_a': {'slope': 1, 'intercept': -2},
        'line_parameter_b': {'slope': -1, 'intercept': -3.},
        'return_value': (-0.5, -2.5),
    },
    3: {
        'line_parameter_a': {'slope': 1, 'intercept': -2},
        'line_parameter_b': {'slope': -1, 'intercept': -3.},
        'return_value': (-0.5, -2.5),
    },
    4: {
        'line_parameter_a': {'slope': 1 / 3, 'intercept': 1.},
        'line_parameter_b': {'slope': -3.0, 'intercept': 11.0},
        'return_value': (3, 2),
    },
    5: {
        'line_parameter_a': {'slope': 1 / 3, 'intercept': 1.},
        'line_parameter_b': {'slope': -3.0, 'intercept': 11.0},
        'return_value': (3, 2),
    },
    6: {
        'line_parameter_a': {'slope': 'VERTICAL', 'intercept': 1419169.19},
        'line_parameter_b':  {'slope': 0.0, 'intercept': 4184658.3044332825},
        'return_value': (1419169.19, 4184658.3044332825)
    },
    7: {
        'line_parameter_a': {'slope': 0.0, 'intercept': 4184658.3044332825},
        'line_parameter_b': {'slope': 'VERTICAL', 'intercept': 1419169.19},
        'return_value': (1419169.19, 4184658.3044332825)
    },
}


def test_all():
    # get_slope_between_two_points
    print(test_function(get_slope_between_two_points, get_slope_between_two_points_parameters))

    # get_intercept_from_point_and_slope
    print(test_function(get_intercept_from_point_and_slope, get_intercept_from_point_and_slope_parameters))

    # get_return_value_from_point_and_intercept_parameters
    print(test_function(get_slope_from_point_and_intercept, get_slope_from_point_and_intercept_parameters))

    # line_parameters
    print(test_function(line_parameters, line_parameters_parameters))

    # perpendicular_line_parameters_at_point
    print(test_function(perpendicular_line_parameters_at_point, perpendicular_line_parameters_at_point_parameters))

    # point_at_distance_with_line_parameters
    print(test_function(point_at_distance_with_line_parameters, point_at_distance_with_line_parameters_parameters))

    # crossing_point_from_lines_parameters
    print(test_function(crossing_point_from_lines_parameters, crossing_point_from_lines_parameters_parameters))


if __name__ == '__main__':
    test_all()



import copy
import pathlib
import types
import warnings
import inspect
import tempfile
import shutil


def test_dependencies():

    try:
        from osgeo import ogr

        import_ogr_success = True
    except ImportError:
        import_ogr_success = False

    try:
        from osgeo import osr

        import_osr_sucess = True
    except ImportError:
        import_osr_sucess = False

    try:
        import psycopg2
        import psycopg2.extras

        import_psycopg2_success = True
    except ImportError:
        import_psycopg2_success = False

    return {
        "ogr": import_ogr_success,
        "osr": import_osr_sucess,
        "psycopg2": import_psycopg2_success,
    }


def print_errors(in_error):
    if in_error:
        sentence = ""
        for error in in_error:
            (id_parameters, error_message, returned_values) = error
            sentence += "\tThe parameters #{} returned errors: \n".format(id_parameters)
            sentence += "\t\t" + error_message + "\n"
            sentence += "\t\t" + returned_values + "\n"
        return sentence
    else:
        return False


def _get_function_value(exec, param):
    try:
        if type(param) is dict:
            result_value = exec(**param)
        else:
            result_value = exec(param)
    except Exception as e:
        result_value = e.__str__()

    return result_value


def _check_path(result_value, return_value):
    if isinstance(return_value, tuple):
        path_in_tuple = False
        if return_value:

            if isinstance(return_value[0], pathlib.Path):
                if len(return_value) == 2 and return_value[1] == "check":
                    if return_value[0].is_file():
                        result_value = return_value
                else:
                    path_in_tuple = True
                    result_value = []
                    for value in return_value:
                        result_value.append(
                            _check_path(result_value=result_value, return_value=value)
                        )
                    result_value = tuple(result_value)

        # if tuple is not composed of path we skip this test
        if path_in_tuple is False:
            return result_value
    else:
        if return_value.is_file():
            result_value = return_value
        else:
            result_value = f"file {return_value} not found"

    return result_value


def _compare_return_value_and_function_value(
    id_parameters, in_error, function, parameters, return_value
):
    # execute function
    function_value = _get_function_value(exec=function, param=parameters)
    in_error = inspect_result_value(
        id_parameters=id_parameters,
        in_error=in_error,
        return_value=return_value,
        function_value=function_value,
    )

    return in_error


def inspect_result_value(id_parameters, in_error, return_value, function_value):

    # if result_value is different from return_value we check specific cases
    if function_value != return_value:
        if isinstance(return_value, (list, tuple)) and isinstance(
            function_value, (list, tuple)
        ):
            for i, i_return_value in enumerate(return_value):
                i_function_value = function_value[i]
                in_error = inspect_result_value(
                    id_parameters, in_error, i_return_value, i_function_value
                )
        else:
            # test return class object
            if inspect.isclass(return_value):
                if isinstance(function_value, return_value):
                    function_value = return_value

            if isinstance(function_value, tempfile._TemporaryFileWrapper):
                with open(function_value.name, "r") as function_data:
                    if function_data.read() == return_value:
                        function_value = return_value

            # if return is a generator we transform it to tuple
            elif isinstance(function_value, types.GeneratorType):
                function_value = _get_function_value(exec=tuple, param=function_value)

            elif isinstance(function_value, zip):
                function_value = _get_function_value(exec=list, param=function_value)

            # TODO FIX IT
            # check path
            # if return value is a path we just verify if file exists
            elif isinstance(return_value, (pathlib.Path, tuple)):
                function_value = _check_path(
                    result_value=function_value, return_value=return_value
                )

            if test_dependencies()["ogr"]:
                from osgeo import ogr

                if isinstance(function_value, ogr.Geometry):
                    if function_value.Equals(return_value):
                        function_value = True

            # exception for ogr geometries
            if test_dependencies()["ogr"]:
                from osgeo import ogr

                if function_value is True:
                    return_value = True

            if function_value != return_value:
                in_error.append(
                    (
                        id_parameters,
                        "ERROR: return value must be {return_value}".format(
                            return_value=return_value
                        ),
                        "ERROR: function return this {result_value}".format(
                            result_value=function_value
                        ),
                    )
                )

    return in_error


def test_function(function, test_parameters):

    in_error = []
    test_parameters_copy = copy.deepcopy(test_parameters)
    for id_parameters, parameters in test_parameters_copy.items():
        return_value = parameters["return_value"]
        del parameters["return_value"]

        in_error = _compare_return_value_and_function_value(
            id_parameters=id_parameters,
            in_error=in_error,
            function=function,
            parameters=parameters,
            return_value=return_value,
        )

    error_value = print_errors(in_error)
    if error_value:
        sentence = "{function_name} KO\n".format(function_name=function.__name__)
        sentence += error_value
        warnings.warn(sentence)
    else:
        sentence = "{function_name} OK".format(function_name=function.__name__)

    return sentence


def create_test_dir(file_path_base, test_dir_path):
    # create test dir
    dir_test_path = file_path_base(test_dir_path)
    if not dir_test_path.exists():
        dir_test_path.mkdir()

    return dir_test_path


def delete_test_dir(dir_test_path):
    shutil.rmtree(path=dir_test_path)



import datetime

from tests.utils.tests_utils import test_function

from geoformat.conversion.datetime_conversion import (
    format_datetime_object_to_str_value,
    date_to_int,
    time_to_int,
    int_to_date,
    int_to_time,
)

from tests.data.features import date_time_value


format_datetime_object_to_str_value_parameters = {
    0: {
        "datetime_value": datetime.datetime(year=2022, month=12, day=12),
        "format": ["year", "month", "day"],
        "return_value": "20221212",
    },
    1: {
        "datetime_value": datetime.datetime(year=2022, month=12, day=12),
        "format": ["year", "month"],
        "return_value": "202212",
    },
    2: {
        "datetime_value": datetime.datetime(year=2022, month=12, day=12, hour=11),
        "format": ["year", "month", "hour"],
        "return_value": "20221211",
    },
}

date_to_int_parameters = {
    0: {
        "date_value": datetime.date(1970, 1, 1),
        "epoch": datetime.date(1970, 1, 1),
        "return_value": 0,
    },
    1: {
        "date_value": datetime.date(1, 1, 1),
        "epoch": datetime.date(1970, 1, 1),
        "return_value": -719162,
    },
    2: {
        "date_value": datetime.date(1970, 1, 1),
        "epoch": datetime.date(1, 1, 1),
        "return_value": 719162,
    },
    3: {
        "date_value": datetime.date(2023, 12, 1),
        "epoch": datetime.date(1970, 1, 1),
        "return_value": 19692,
    },
    4: {
        "date_value": datetime.date(2023, 12, 31),
        "epoch": datetime.date(1970, 1, 1),
        "return_value": 19692+30,
    },
    5: {
        "date_value": datetime.date(2011, 2, 10),
        "epoch": datetime.date(1970, 1, 1),
        "return_value": 15015,
    },
}


time_to_int_parameters = {
    0: {
        "time_value": date_time_value.time(),
        "return_value": 40930000999
    },
    1: {
        "time_value": datetime.time(hour=0, minute=0, second=0, microsecond=0),
        "return_value": 0
    },
    2: {
        "time_value": datetime.time(hour=0, minute=0, second=0, microsecond=1),
        "return_value": 1
    },
    3: {
        "time_value": datetime.time(hour=23, minute=59, second=59, microsecond=999999),
        "return_value": 86399999999
    },
}


int_to_date_parameters = {
    0: {
        "int_value": 0,
        "epoch": datetime.date(1970, 1, 1),
        "return_value": datetime.date(1970, 1, 1),
    },
    1: {
        "int_value": -719162,
        "epoch": datetime.date(1970, 1, 1),
        "return_value": datetime.date(1, 1, 1),
    },
    2: {
        "int_value": 719162,
        "epoch": datetime.date(1, 1, 1),
        "return_value": datetime.date(1970, 1, 1),
    },
    3: {
        "int_value": 19692,
        "epoch": datetime.date(1970, 1, 1),
        "return_value": datetime.date(2023, 12, 1),
    },
    4: {
        "int_value": 19692+30,
        "epoch": datetime.date(1970, 1, 1),
        "return_value": datetime.date(2023, 12, 31),
    },
    5: {
        "int_value": 15015,
        "epoch": datetime.date(1970, 1, 1),
        "return_value": datetime.date(2011, 2, 10),
    },
}

int_to_time_parameters = {
    0: {
        "int_value": 40930000999,
        "return_value": date_time_value.time()
    },
    1: {
        "int_value": 0,
        "return_value": datetime.time(hour=0, minute=0, second=0, microsecond=0)
    },
    2: {
        "int_value": 1,
        "return_value": datetime.time(hour=0, minute=0, second=0, microsecond=1)
    },
    3: {
        "int_value": 86399999999,
        "return_value": datetime.time(hour=23, minute=59, second=59, microsecond=999999)
    },
}


def test_all():
    # format_datetime_object_to_str_value
    print(
        test_function(
            format_datetime_object_to_str_value,
            format_datetime_object_to_str_value_parameters,
        )
    )

    print(test_function(date_to_int, date_to_int_parameters))

    print(test_function(time_to_int, time_to_int_parameters))

    print(test_function(int_to_date, int_to_date_parameters))

    print(test_function(int_to_time, int_to_time_parameters))


if __name__ == "__main__":
    test_all()



from geoformat.index.attributes.hash import create_attribute_index

from tests.data.geolayers import geolayer_fr_dept_population
from tests.data.index import geolayer_fr_dept_population_CODE_DEPT_hash_index

from tests.utils.tests_utils import test_function

create_attribute_index_parameters = {
    0: {
        "geolayer": geolayer_fr_dept_population,
        "field_name": "CODE_DEPT",
        "return_value": geolayer_fr_dept_population_CODE_DEPT_hash_index
    },
    1: {
        "geolayer": geolayer_fr_dept_population,
        "field_name": "INSEE_REG",
        "return_value": {'type': 'hashtable', 'index': {'76': [0, 8, 28, 31, 36, 40, 42, 47, 48, 57, 77, 79, 92], '75': [1, 13, 16, 22, 25, 26, 32, 56, 64, 75, 83, 87], '84': [2, 6, 20, 30, 35, 44, 55, 66, 74, 76, 78, 80], '32': [3, 21, 46, 59, 67], '44': [4, 5, 15, 18, 24, 39, 41, 45, 54, 58], '93': [7, 49, 82, 84, 89, 93], '27': [9, 14, 27, 34, 60, 62, 69, 72], '52': [10, 29, 50, 61, 88], '11': [11, 43, 68, 71, 90, 91, 94, 95], '28': [12, 17, 37, 38, 70], '24': [19, 23, 33, 53, 81, 85], '53': [51, 63, 65, 73], '94': [52, 86]}}
    }
}

def test_all():

    # create_attribute_index
    print(test_function(create_attribute_index, create_attribute_index_parameters))

if __name__ == '__main__':
    test_all()


from tests.data.fields_metadata import (
    geolayer_data_fields_metadata_complete,
    geolayer_attributes_only_fields_metadata,
)

metadata_fr_dept_data_and_geometry = {
    "name": "FRANCE_DPT_GENERALIZE_LAMB93_ROUND_DATA_AND_GEOMETRY",
    "fields": geolayer_data_fields_metadata_complete,
    "geometry_ref": {"type": {"MultiPolygon", "Polygon"}, "crs": 2154},
}

metadata_fr_dept_data_only = {
    "name": "FRANCE_DPT_GENERALIZE_LAMB93_ROUND_DATA_ONLY",
    "fields": geolayer_data_fields_metadata_complete,
}

metadata_fr_dept_geometry_only = {
    "name": "FRANCE_DPT_GENERALIZE_LAMB93_ROUND_GEOMETRY_ONLY",
    "geometry_ref": {
        "type": {"MultiPolygon", "Polygon"},
        "extent": (124277.0, 6050136.0, 1242213.0, 7110430.0),
        "crs": 2154,
    },
}

metadata_fr_dept_population = {
    "name": "dept_population",
    "fields": {
        "CODE_DEPT": {"type": "String", "width": 2, "index": 0},
        "INSEE_REG": {"type": "String", "width": 2, "index": 1},
        "POPULATION": {"type": "Integer", "index": 2},
        "AREA": {"type": "Real", "width": 7, "precision": 2, "index": 3},
        "DENSITY": {"type": "Real", "width": 7, "precision": 2, "index": 4},
    },
}

metadata_fr_dept_population_geometry = {
    "name": "FRANCE_DPT_WITH_POPULATION",
    "fields": {
        "CODE_DEPT": {"type": "String", "width": 2, "index": 0},
        "NOM_DEPT": {"type": "String", "width": 23, "index": 1},
        "POPULATION": {"type": "Integer", "index": 2},
        "DENSITY": {"type": "Real", "width": 7, "precision": 2, "index": 3},
    },
    "geometry_ref": {"type": {"Polygon", "MultiPolygon"}, "crs": 2154},
}

metadata_fr_dept_population_geometry_right = {
    "name": "fr_dept_data_and_geometry_right_join_fr_dept_population",
    "fields": {
        "POPULATION": {"type": "Integer", "index": 0},
        "DENSITY": {"type": "Real", "width": 7, "precision": 2, "index": 1},
        "CODE_DEPT": {"type": "String", "width": 2, "index": 2},
        "NOM_DEPT": {"type": "String", "width": 23, "index": 3},
    },
    "geometry_ref": {"type": {"Polygon", "MultiPolygon"}, "crs": 2154},
}

metadata_paris_velib = {
    "name": "geolayer_paris_velib",
    "fields": {
        "Identifiant station": {"type": "Integer", "index": 0},
        "Nom station": {"type": "String", "width": 45, "index": 1},
        "Station en fonctionnement": {"type": "String", "width": 3, "index": 2},
        "Capacité de la station": {"type": "Integer", "index": 3},
        "Nombre bornettes libres": {"type": "Integer", "index": 4},
        "Nombre total vélos disponibles": {"type": "Integer", "index": 5},
        "Vélos mécaniques disponibles": {"type": "Integer", "index": 6},
        "Vélos électriques disponibles": {"type": "Integer", "index": 7},
        "Borne de paiement disponible": {"type": "String", "width": 3, "index": 8},
        "Retour vélib possible": {"type": "String", "width": 3, "index": 9},
        "Actualisation de la donnée": {"type": "DateTime", "index": 10},
        "Coordonnées géographiques": {
            "type": "RealList",
            "width": 13,
            "precision": 11,
            "index": 11,
        },
        "Nom communes équipées": {"type": "String", "width": 20, "index": 12},
    },
}

metadata_attributes_to_force_in_str = {
    "name": "attributes_to_force_only_forced",
    "fields": {
        "field_integer": {"type": "Real", "width": 9, "precision": 5, "index": 0},
        "field_integer_list": {"type": "String", "width": 13, "index": 1},
        "field_real": {"type": "String", "width": 7, "index": 2},
        "field_real_list": {"type": "String", "width": 19, "index": 3},
        "field_string": {"type": "String", "width": 26, "index": 4},
        "field_date": {"type": "Date", "index": 5},
        "field_time": {"type": "String", "width": 15, "index": 6},
        "field_binary": {"type": "String", "width": 201, "index": 7},
        "field_boolean": {"type": "String", "width": 5, "index": 8},
        "field_string_list": {"type": "String", "width": 16, "index": 9},
        "field_datetime": {"type": "DateTime", "index": 10},
    },
}

metadata_attributes_only = {
    "name": "attributes_only",
    "fields": geolayer_attributes_only_fields_metadata,
}

metadata_geometry_only_all_geometries_type = {
    "name": "all_geometry_type_only",
    "geometry_ref": {
        "type": {
            "Point",
            "LineString",
            "Polygon",
            "MultiPoint",
            "MultiLineString",
            "MultiPolygon",
            "GeometryCollection",
        },
        "crs": 4326,
    },
}

metadata_geometry_2d = {
    "name": 'geolayer_geometry',
    "geometry_ref": {"type": {"GeometryCollection"}}
}

metadata_geometry_3d = {
    "name": 'geolayer_geometry',
    "geometry_ref": {"type": {"GeometryCollection25D"}}
}

metadata_paris_tokyo = {
    "name": "paris_tokyo",
    "fields": {
        "name": {"type": "String", "width": 5, "index": 0},
        "country": {"type": "String", "width": 6, "index": 1},
    },
    "geometry_ref": {
        "type": {"Point"},
    },
}


import datetime

from tests.utils.tests_utils import test_function

from geoformat.conf.error_messages import field_width_float_not_none

from geoformat.conversion.metadata_conversion import (
    geometries_scan_to_geometries_metadata,
    fields_scan_to_fields_metadata,
    get_field_name_list_ordered_by_i_field,
    from_field_scan_determine_field_width,
)

from tests.data.fields_metadata import (
    geolayer_attributes_only_fields_metadata,
    geolayer_attributes_only_fields_metadata_without_index,
)

geometries_scan_to_geometries_metadata_parameters = {
    0: {
        "geometry_scan": {"type": {"Polygon", "MultiPolygon"}, "extent": None},
        "extent": False,
        "crs": None,
        "return_value": {"type": {"Polygon", "MultiPolygon"}},
    },
    1: {
        "geometry_scan": {"type": {"Polygon", "MultiPolygon"}, "extent": None},
        "extent": True,
        "crs": None,
        "return_value": {"type": {"Polygon", "MultiPolygon"}},
    },
    2: {
        "geometry_scan": {"type": None, "extent": None},
        "extent": False,
        "crs": None,
        "return_value": None,
    },
    3: {
        "geometry_scan": {"type": None, "extent": None},
        "extent": True,
        "crs": None,
        "return_value": None,
    },
    4: {
        "geometry_scan": {"type": {"Polygon"}, "extent": None},
        "crs": None,
        "extent": True,
        "return_value": {"type": {"Polygon"}},
    },
    5: {
        "geometry_scan": {
            "type": {
                "Point",
                "LineString",
                "Polygon",
                "MultiPoint",
                "MultiLineString",
                "MultiPolygon",
                "GeometryCollection",
            },
            "extent": None,
        },
        "crs": None,
        "extent": True,
        "return_value": {
            "type": {
                "Point",
                "LineString",
                "Polygon",
                "MultiPoint",
                "MultiLineString",
                "MultiPolygon",
                "GeometryCollection",
            }
        },
    },
    6: {
        "geometry_scan": {"type": {"Polygon", "MultiPolygon"}, "extent": None},
        "crs": None,
        "extent": True,
        "return_value": {"type": {"Polygon", "MultiPolygon"}},
    },
    7: {
        "geometry_scan": {
            "type": {"Polygon", "MultiPolygon"},
            "extent": (124277.0, 6050136.0, 1242213.0, 7110430.0),
        },
        "crs": None,
        "extent": True,
        "return_value": {
            "type": {"Polygon", "MultiPolygon"},
            "extent": (124277.0, 6050136.0, 1242213.0, 7110430.0),
        },
    },
    8: {
        "geometry_scan": {"type": None, "extent": None},
        "crs": None,
        "extent": True,
        "return_value": None,
    },
    9: {
        "geometry_scan": {
            "type": {"Polygon"},
            "extent": (382932.0, 6743442.0, 790134.0, 6997000.0),
        },
        "crs": None,
        "extent": True,
        "return_value": {
            "type": {"Polygon"},
            "extent": (382932.0, 6743442.0, 790134.0, 6997000.0),
        },
    },
    10: {
        "geometry_scan": {
            "type": {
                "Point",
                "LineString",
                "Polygon",
                "MultiPoint",
                "MultiLineString",
                "MultiPolygon",
                "GeometryCollection",
            },
            "extent": (-157.97, -34.29, 35.12, 77.95),
        },
        "crs": None,
        "extent": True,
        "return_value": {
            "type": {
                "Point",
                "LineString",
                "Polygon",
                "MultiPoint",
                "MultiLineString",
                "MultiPolygon",
                "GeometryCollection",
            },
            "extent": (-157.97, -34.29, 35.12, 77.95),
        },
    },
    11: {
        "geometry_scan": {
            "type": {"Polygon", "MultiPolygon"},
            "extent": (124277.0, 6050136.0, 1242213.0, 7110430.0),
        },
        "extent": True,
        "crs": 2154,
        "return_value": {
            "type": {"Polygon", "MultiPolygon"},
            "extent": (124277.0, 6050136.0, 1242213.0, 7110430.0),
            "crs": 2154,
        },
    },
    12: {
        "geometry_scan": {
            "type": {
                "Point",
                "LineString",
                "Polygon",
                "MultiPoint",
                "MultiLineString",
                "MultiPolygon",
                "GeometryCollection",
            },
            "extent": (-157.97, -34.29, 35.12, 77.95),
        },
        "extent": True,
        "crs": 4326,
        "return_value": {
            "type": {
                "Point",
                "LineString",
                "Polygon",
                "MultiPoint",
                "MultiLineString",
                "MultiPolygon",
                "GeometryCollection",
            },
            "extent": (-157.97, -34.29, 35.12, 77.95),
            "crs": 4326,
        },
    },
}

fields_scan_to_fields_metadata_parameters = {
    0: {
        "fields_scan": {
            "field_integer": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": float,
                "field_width_str": 6,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 0,
                "field_recast": False,
                "native_type": {float},
                "str_is_always_hexadecimal": None,
                "force_type": False,
                "field_width_float": 4,
                "width_before_comma": 4,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_integer_list": {
                "values_in_list": True,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 12,
                "field_width_list": 13,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 1,
                "field_recast": True,
                "native_type": {str, int, float},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 4,
                "width_before_comma": 4,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_real": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 7,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 2,
                "field_recast": True,
                "native_type": {str, int},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_real_list": {
                "values_in_list": True,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 19,
                "field_width_list": 6,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 3,
                "field_recast": True,
                "native_type": {str, int},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_string": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 26,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_index": 4,
                "field_recast": True,
                "native_type": {str, datetime.datetime},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_date": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.date,
                "field_width_str": 10,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_index": 5,
                "field_recast": False,
                "native_type": {datetime.date},
                "str_is_always_hexadecimal": None,
                "force_type": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_time": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 15,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 6,
                "field_recast": True,
                "native_type": {str, datetime.time},
                "str_is_always_hexadecimal": False,
                "force_type": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_binary": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 201,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 7,
                "field_recast": True,
                "native_type": {str, bytes},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_boolean": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 5,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 8,
                "field_recast": True,
                "native_type": {bool, int},
                "str_is_always_hexadecimal": None,
                "force_type": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_none": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 4,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_index": 9,
                "field_recast": False,
                "native_type": {str},
                "str_is_always_hexadecimal": False,
                "force_type": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_string_list": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 16,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_index": 10,
                "field_recast": False,
                "native_type": {str},
                "str_is_always_hexadecimal": False,
                "force_type": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_datetime": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.datetime,
                "field_width_str": 26,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_index": 11,
                "field_recast": False,
                "native_type": {datetime.datetime},
                "str_is_always_hexadecimal": None,
                "force_type": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
        },
        "return_value": {
            "field_integer": {"type": "Real", "width": 4, "precision": 0, "index": 0},
            "field_integer_list": {"type": "String", "width": 13, "index": 1},
            "field_real": {"type": "String", "width": 7, "index": 2},
            "field_real_list": {"type": "String", "width": 19, "index": 3},
            "field_string": {"type": "String", "width": 26, "index": 4},
            "field_date": {"type": "Date", "index": 5},
            "field_time": {"type": "String", "width": 15, "index": 6},
            "field_binary": {"type": "String", "width": 201, "index": 7},
            "field_boolean": {"type": "String", "width": 5, "index": 8},
            "field_none": {"type": "String", "width": 4, "index": 9},
            "field_string_list": {"type": "String", "width": 16, "index": 10},
            "field_datetime": {"type": "DateTime", "index": 11},
        },
    },
    1: {
        "fields_scan": {
            "field_integer": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": float,
                "field_width_str": 6,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": False,
                "native_type": {float},
                "str_is_always_hexadecimal": None,
                "force_type": False,
                "field_width_float": 4,
                "width_before_comma": 4,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_integer_list": {
                "values_in_list": True,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 12,
                "field_width_list": 13,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {str, int, float},
                "str_is_always_hexadecimal": False,
                "force_type": False,
                "field_width_float": 4,
                "width_before_comma": 4,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_real": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 7,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {str, int},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_real_list": {
                "values_in_list": True,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 19,
                "field_width_list": 6,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {str, int},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_string": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 26,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {str, datetime.datetime},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_date": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.date,
                "field_width_str": 10,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_recast": False,
                "native_type": {datetime.date},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_time": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 15,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {str, datetime.time},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_binary": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 201,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {str, bytes},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_boolean": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 5,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {bool, int},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_none": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 4,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_recast": False,
                "native_type": {str},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_string_list": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 16,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_recast": False,
                "native_type": {str},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_datetime": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.datetime,
                "field_width_str": 26,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_recast": False,
                "native_type": {datetime.datetime},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
        },
        "return_value": {
            "field_integer": {"type": "Real", "width": 4, "precision": 0},
            "field_integer_list": {"type": "String", "width": 13},
            "field_real": {"type": "String", "width": 7},
            "field_real_list": {"type": "String", "width": 19},
            "field_string": {"type": "String", "width": 26},
            "field_date": {"type": "Date"},
            "field_time": {"type": "String", "width": 15},
            "field_binary": {"type": "String", "width": 201},
            "field_boolean": {"type": "String", "width": 5},
            "field_none": {"type": "String", "width": 4},
            "field_string_list": {"type": "String", "width": 16},
            "field_datetime": {"type": "DateTime"},
        },
    },
    2: {
        "fields_scan": {
            "field_integer": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": int,
                "field_width_str": 3,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 0,
                "field_recast": False,
                "native_type": {int},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                "field_width_float": 3,
                "width_before_comma": 3,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_integer_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": int,
                "field_width_str": 4,
                "field_width_list": 12,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 1,
                "field_recast": False,
                "native_type": {int},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                "field_width_float": 4,
                "width_before_comma": 4,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_real": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": int,
                "field_width_str": 6,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 2,
                "field_recast": True,
                "native_type": {float},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                "field_width_float": 4,
                "width_before_comma": 4,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_real_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": float,
                "field_width_str": 10,
                "field_width_list": 24,
                "field_precision": 5,
                "none_value": False,
                "not_none_value": True,
                "field_index": 3,
                "field_recast": False,
                "native_type": {float},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                "field_width_float": 10,
                "width_before_comma": 5,
                "width_after_comma": 5,
                "field_delete": False,
            },
            "field_string": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 5,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 4,
                "field_recast": False,
                "native_type": {str},
                "force_type": True,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_string_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": str,
                "field_width_str": 8,
                "field_width_list": 23,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 5,
                "field_recast": False,
                "native_type": {str},
                "force_type": True,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_none": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 6,
                "field_recast": False,
                "native_type": set(),
                "force_type": True,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": True,
            },
            "field_date": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.date,
                "field_width_str": 10,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 7,
                "field_recast": False,
                "native_type": {datetime.date},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_time": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.time,
                "field_width_str": 15,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 8,
                "field_recast": False,
                "native_type": {datetime.time},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_datetime": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.datetime,
                "field_width_str": 26,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 9,
                "field_recast": False,
                "native_type": {datetime.datetime},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_binary": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bytes,
                "field_width_str": 201,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 10,
                "field_recast": False,
                "native_type": {bytes},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_boolean": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bool,
                "field_width_str": 4,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 11,
                "field_recast": False,
                "native_type": {bool},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
        },
        "return_value": {
            "field_integer": {"type": "Integer", "index": 0},
            "field_integer_list": {"type": "IntegerList", "index": 1},
            "field_real": {"type": "Integer", "index": 2},
            "field_real_list": {
                "type": "RealList",
                "width": 10,
                "precision": 5,
                "index": 3,
            },
            "field_string": {"type": "String", "width": 5, "index": 4},
            "field_string_list": {"type": "StringList", "width": 8, "index": 5},
            "field_date": {"type": "Date", "index": 6},
            "field_time": {"type": "Time", "index": 7},
            "field_datetime": {"type": "DateTime", "index": 8},
            "field_binary": {"type": "Binary", "index": 9},
            "field_boolean": {"type": "Boolean", "index": 10},
        },
    },
    3: {
        "fields_scan": {
            "field_integer": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": int,
                "field_width_str": 3,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 0,
                "field_recast": False,
                "native_type": {int},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_integer_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": int,
                "field_width_str": 4,
                "field_width_list": 12,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 1,
                "field_recast": False,
                "native_type": {int},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_real": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": float,
                "field_width_str": 6,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 2,
                "field_recast": False,
                "native_type": {float},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                "field_width_float": 4,
                "width_before_comma": 4,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_real_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": float,
                "field_width_str": 10,
                "field_width_list": 24,
                "field_precision": 5,
                "none_value": False,
                "not_none_value": True,
                "field_index": 3,
                "field_recast": False,
                "native_type": {float},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                "field_width_float": 10,
                "width_before_comma": 5,
                "width_after_comma": 5,
                "field_delete": False,
            },
            "field_string": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 5,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 4,
                "field_recast": False,
                "native_type": {str},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_string_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": str,
                "field_width_str": 8,
                "field_width_list": 23,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 5,
                "field_recast": False,
                "native_type": {str},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_none": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 6,
                "field_recast": False,
                "native_type": set(),
                "force_type": False,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": True,
            },
            "field_date": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.date,
                "field_width_str": 10,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 7,
                "field_recast": False,
                "native_type": {datetime.date},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_time": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.time,
                "field_width_str": 15,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 8,
                "field_recast": False,
                "native_type": {datetime.time},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_datetime": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.datetime,
                "field_width_str": 26,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 9,
                "field_recast": False,
                "native_type": {datetime.datetime},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_binary": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bytes,
                "field_width_str": 201,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 10,
                "field_recast": False,
                "native_type": {bytes},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_boolean": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bool,
                "field_width_str": 4,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 11,
                "field_recast": False,
                "native_type": {bool},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
        },
        "return_value": {
            "field_integer": {"type": "Integer", "index": 0},
            "field_integer_list": {"type": "IntegerList", "index": 1},
            "field_real": {"type": "Real", "width": 4, "precision": 0, "index": 2},
            "field_real_list": {
                "type": "RealList",
                "width": 10,
                "precision": 5,
                "index": 3,
            },
            "field_string": {"type": "String", "width": 5, "index": 4},
            "field_string_list": {"type": "StringList", "width": 8, "index": 5},
            "field_date": {"type": "Date", "index": 6},
            "field_time": {"type": "Time", "index": 7},
            "field_datetime": {"type": "DateTime", "index": 8},
            "field_binary": {"type": "Binary", "index": 9},
            "field_boolean": {"type": "Boolean", "index": 10},
        },
    },
    4: {
        "fields_scan": {
            "field_none": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 0,
                "field_recast": False,
                "native_type": set(),
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": True,
            },
            "field_integer": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": int,
                "field_width_str": 3,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 1,
                "field_recast": False,
                "native_type": {int},
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_integer_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": int,
                "field_width_str": 4,
                "field_width_list": 12,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 2,
                "field_recast": False,
                "native_type": {int},
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_real": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": int,
                "field_width_str": 6,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 3,
                "field_recast": True,
                "native_type": {float},
                "force_type": True,
                "field_width_float": 4,
                "width_before_comma": 4,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_real_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": float,
                "field_width_str": 10,
                "field_width_list": 24,
                "field_precision": 5,
                "none_value": False,
                "not_none_value": True,
                "field_index": 4,
                "field_recast": False,
                "native_type": {float},
                "force_type": True,
                "field_width_float": 10,
                "width_before_comma": 5,
                "width_after_comma": 5,
                "field_delete": False,
            },
            "field_string": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 5,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 5,
                "field_recast": False,
                "native_type": {str},
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_string_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": str,
                "field_width_str": 8,
                "field_width_list": 23,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 6,
                "field_recast": False,
                "native_type": {str},
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_none_1": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 7,
                "field_recast": False,
                "native_type": set(),
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": True,
            },
            "field_date": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.date,
                "field_width_str": 10,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 8,
                "field_recast": False,
                "native_type": {datetime.date},
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_time": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.time,
                "field_width_str": 15,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 9,
                "field_recast": False,
                "native_type": {datetime.time},
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_none_2": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 10,
                "field_recast": False,
                "native_type": set(),
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": True,
            },
            "field_none_3": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 11,
                "field_recast": False,
                "native_type": set(),
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": True,
            },
            "field_datetime": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.datetime,
                "field_width_str": 26,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 12,
                "field_recast": False,
                "native_type": {datetime.datetime},
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_binary": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bytes,
                "field_width_str": 201,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 13,
                "field_recast": False,
                "native_type": {bytes},
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_boolean": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bool,
                "field_width_str": 4,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 14,
                "field_recast": False,
                "native_type": {bool},
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            },
            "field_none_4": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 15,
                "field_recast": False,
                "native_type": set(),
                "force_type": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": True,
            },
        },
        "return_value": {
            "field_integer": {"type": "Integer", "index": 0},
            "field_integer_list": {"type": "IntegerList", "index": 1},
            "field_real": {"type": "Integer", "index": 2},
            "field_real_list": {
                "type": "RealList",
                "width": 10,
                "precision": 5,
                "index": 3,
            },
            "field_string": {"type": "String", "width": 5, "index": 4},
            "field_string_list": {"type": "StringList", "width": 8, "index": 5},
            "field_date": {"type": "Date", "index": 6},
            "field_time": {"type": "Time", "index": 7},
            "field_datetime": {"type": "DateTime", "index": 8},
            "field_binary": {"type": "Binary", "index": 9},
            "field_boolean": {"type": "Boolean", "index": 10},
        },
    },
    5: {
        "fields_scan": {
            "field_binary": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bytes,
                "field_width_str": 201,
                "field_width_list": 0,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 0,
                "field_recast": True,
                "native_type": {str, bytes},
                "force_type": True,
                "str_is_always_hexadecimal": True,
                "field_width_float": 0,
                "width_before_comma": 0,
                "width_after_comma": 0,
                "field_delete": False,
            }
        },
        "return_value": {"field_binary": {"type": "Binary", "index": 0}},
    },
}

get_field_name_list_ordered_by_i_field_parameters = {
    0: {
        "fields_metadata": geolayer_attributes_only_fields_metadata,
        "return_value": [
            "field_integer",
            "field_integer_list",
            "field_real",
            "field_real_list",
            "field_string",
            "field_string_list",
            "field_date",
            "field_time",
            "field_datetime",
            "field_binary",
            "field_boolean",
        ],
    },
    1: {
        "fields_metadata": geolayer_attributes_only_fields_metadata_without_index,
        "return_value": [
            "field_integer",
            "field_integer_list",
            "field_real",
            "field_real_list",
            "field_string",
            "field_string_list",
            "field_date",
            "field_time",
            "field_datetime",
            "field_binary",
            "field_boolean",
        ],
    },
    2: {
        "fields_metadata": {
            "commune": {"type": "String", "width": 254, "index": 2},
            "id_uic": {"type": "Real", "width": 19, "precision": 11, "index": 0},
            "id_ligne": {"type": "String", "width": 254, "index": 1},
            "dpt": {"type": "String", "width": 254, "index": 3},
        },
        "return_value": ["id_uic", "id_ligne", "commune", "dpt"],
    },
}

from_field_scan_determine_field_width_parameters = {
    0: {
        "field_type": str,
        "field_in_list": False,
        "field_width_str": 4,
        "field_width_float": None,
        "field_width_list": None,
        "max_width": None,
        "return_value": 4,
    },
    1: {
        "field_type": str,
        "field_in_list": False,
        "field_width_str": 4,
        "field_width_float": None,
        "field_width_list": 20,
        "max_width": None,
        "return_value": 20,
    },
    2: {
        "field_type": str,
        "field_in_list": False,
        "field_width_str": 4,
        "field_width_float": 25,
        "field_width_list": 20,
        "max_width": None,
        "return_value": 20,
    },
    3: {
        "field_type": str,
        "field_in_list": False,
        "field_width_str": 20,
        "field_width_float": 25,
        "field_width_list": 4,
        "max_width": None,
        "return_value": 20,
    },
    4: {
        "field_type": str,
        "field_in_list": True,
        "field_width_str": 4,
        "field_width_float": None,
        "field_width_list": None,
        "max_width": None,
        "return_value": 4,
    },
    5: {
        "field_type": str,
        "field_in_list": True,
        "field_width_str": 4,
        "field_width_float": None,
        "field_width_list": 20,
        "max_width": None,
        "return_value": 4,
    },
    6: {
        "field_type": str,
        "field_in_list": True,
        "field_width_str": 4,
        "field_width_float": 25,  # not possible in reality but only for testing.
        "field_width_list": 20,
        "max_width": None,
        "return_value": 4,
    },
    7: {
        "field_type": str,
        "field_in_list": True,
        "field_width_str": 20,
        "field_width_float": 25,  # not possible in reality but only for testing.
        "field_width_list": 4,  # not possible in reality but only for testing.
        "max_width": None,
        "return_value": 20,
    },
    9: {
        "field_type": float,
        "field_in_list": False,
        "field_width_str": 4,
        "field_width_float": None,
        "field_width_list": None,
        "max_width": None,
        "return_value": field_width_float_not_none,
    },
    10: {
        "field_type": float,
        "field_in_list": False,
        "field_width_str": 4,
        "field_width_float": None,
        "field_width_list": 20,
        "max_width": None,
        "return_value": field_width_float_not_none,
    },
    11: {
        "field_type": float,
        "field_in_list": False,
        "field_width_str": 4,
        "field_width_float": 25,
        "field_width_list": 20,
        "max_width": None,
        "return_value": 25,
    },
    12: {
        "field_type": float,
        "field_in_list": False,
        "field_width_str": 20,
        "field_width_float": 25,
        "field_width_list": 4,
        "max_width": None,
        "return_value": 25,
    },
    13: {
        "field_type": float,
        "field_in_list": True,
        "field_width_str": 4,
        "field_width_float": None,
        "field_width_list": None,
        "max_width": None,
        "return_value": field_width_float_not_none,
    },
    14: {
        "field_type": float,
        "field_in_list": True,
        "field_width_str": 4,
        "field_width_float": None,
        "field_width_list": 20,
        "max_width": None,
        "return_value": field_width_float_not_none,
    },
    15: {
        "field_type": float,
        "field_in_list": True,
        "field_width_str": 4,
        "field_width_float": 25,  # not possible in reality but only for testing.
        "field_width_list": 20,
        "max_width": None,
        "return_value": 25,
    },
    16: {
        "field_type": float,
        "field_in_list": True,
        "field_width_str": 20,
        "field_width_float": 25,  # not possible in reality but only for testing.
        "field_width_list": 4,  # not possible in reality but only for testing.
        "max_width": None,
        "return_value": 25,
    },
    17: {
        "field_type": str,
        "field_in_list": False,
        "field_width_str": 20,
        "field_width_float": 25,  # not possible in reality but only for testing.
        "field_width_list": 4,  # not possible in reality but only for testing.
        "max_width": 10,
        "return_value": 10,
    },
    18: {
        "field_type": float,
        "field_in_list": False,
        "field_width_str": 20,
        "field_width_float": 25,  # not possible in reality but only for testing.
        "field_width_list": 4,  # not possible in reality but only for testing.
        "max_width": 10,
        "return_value": 10,
    },
    19: {
        "field_type": str,
        "field_in_list": True,
        "field_width_str": 20,
        "field_width_float": 25,  # not possible in reality but only for testing.
        "field_width_list": 4,  # not possible in reality but only for testing.
        "max_width": 10,
        "return_value": 10,
    },
    20: {
        "field_type": float,
        "field_in_list": True,
        "field_width_str": 20,
        "field_width_float": 25,  # not possible in reality but only for testing.
        "field_width_list": 4,  # not possible in reality but only for testing.
        "max_width": 10,
        "return_value": 10,
    },
}


def test_all():
    # geometries_scan_to_metadata
    print(
        test_function(
            geometries_scan_to_geometries_metadata,
            geometries_scan_to_geometries_metadata_parameters,
        )
    )

    # fields_scan_to_fields_metadata
    print(
        test_function(
            fields_scan_to_fields_metadata, fields_scan_to_fields_metadata_parameters
        )
    )

    # get_field_name_list_ordered_by_i_field
    print(
        test_function(
            get_field_name_list_ordered_by_i_field,
            get_field_name_list_ordered_by_i_field_parameters,
        )
    )

    # from_field_scan_determine_field_width_parameters
    print(
        test_function(
            from_field_scan_determine_field_width,
            from_field_scan_determine_field_width_parameters,
        )
    )


if __name__ == "__main__":
    test_all()
